package hpdos_rdma_offloaded.service;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;

import com.ibm.darpc.DaRPCClientEndpoint;
import com.ibm.darpc.DaRPCClientGroup;
import com.ibm.darpc.DaRPCServerEndpoint;
import com.ibm.darpc.DaRPCServerEvent;
import com.ibm.darpc.DaRPCService;
import com.ibm.darpc.DaRPCStream;

import org.rocksdb.RocksDBException;

import hpdos_rdma_offloaded.handler.NetworkHandler;
import hpdos_rdma_offloaded.handler.StorageHandler;
import hpdos_rdma_offloaded.lib.Packet;
import hpdos_rdma_offloaded.protocol.AckType;
import hpdos_rdma_offloaded.protocol.InvalidationRequest;
import hpdos_rdma_offloaded.protocol.InvalidationResponse;
import hpdos_rdma_offloaded.protocol.InvalidationRpcProtocol;
import hpdos_rdma_offloaded.protocol.Request;
import hpdos_rdma_offloaded.protocol.RequestType;
import hpdos_rdma_offloaded.protocol.Response;
import hpdos_rdma_offloaded.protocol.RpcProtocol;

public class MasterMetadataServerService extends RpcProtocol implements DaRPCService<Request, Response>{
    NetworkHandler networkHandler;
    StorageHandler storageHandler;
    ConcurrentHashMap<SocketAddress,DaRPCStream<InvalidationRequest, InvalidationResponse>> invalidationClientEpStreams;
    ConcurrentHashMap<SocketAddress,DaRPCClientEndpoint<InvalidationRequest, InvalidationResponse>> invalidationClientEps;
    DaRPCClientGroup<InvalidationRequest, InvalidationResponse> invalidationClientGroup;
    public MasterMetadataServerService(NetworkHandler networkHandler,StorageHandler storageHandler) throws Exception{
        System.out.println("New MetadataServerService Created");
        this.networkHandler = networkHandler;
        this.storageHandler  = storageHandler;
        this.invalidationClientEpStreams = new ConcurrentHashMap<>();
        this.invalidationClientEps = new ConcurrentHashMap<>();
        this.networkHandler.setClientStreams(this.invalidationClientEpStreams);
        InvalidationRpcProtocol rpcProtocol = new InvalidationRpcProtocol();
        invalidationClientGroup  =  DaRPCClientGroup.createClientGroup(rpcProtocol, 100, 0, 16, 16);  
    }

    @Override
    public void processServerEvent(DaRPCServerEvent<Request, Response> event) throws IOException {
        Request request = event.getReceiveMessage();
        Response response = event.getSendMessage();
        Packet packet = new Packet(request.getRequestType(), request.getKey(), request.getValue());
        System.out.print("Received "+request.getRequestType()+" Request for key: " + new String(request.getKey()));
        if(request.getValue() != null)
        System.out.print(" value : "+ new String(request.getValue()) );
        System.out.println();
        try{
            if (request.getRequestType() == RequestType.PUT) {
                this.storageHandler.create(request);
                this.networkHandler.create(packet);
                this.networkHandler.sendInvalidateRequest(request);
               //Yet to consume isDone 
                // var isDone  = this.networkHandler.consumeReplicationRequestFutures(replicationRequestFutures);

                response.setAck(AckType.SUCCESS);
                response.setKey(null);
                response.setValue(null);
            } 
            else if(request.getRequestType() == RequestType.GET) {
                byte[] value = this.storageHandler.read(request); 
                if(value == null)
                {
                    response.setAck(AckType.NOTFOUND);
                }
                else
                {
                    response.setAck(AckType.SUCCESS_WITH_VALUE);
                    response.setValue(value);
                }
            }
            else if(request.getRequestType() == RequestType.DELETE){
                this.storageHandler.delete(request);
                // var replicationRequestFutures = this.networkHandler.sendReplicateRequest(request);
               //Yet to consume isDone 
                // var isDone  = this.networkHandler.consumeReplicationRequestFutures(replicationRequestFutures);
                response.setAck(AckType.SUCCESS);
            }
            else{
                response.setAck(AckType.INVALID_REQUEST);
            }
        }
        catch(RocksDBException | InterruptedException | ExecutionException ex){
                ex.printStackTrace();
                response.setAck(AckType.DBFAILED);
        }
        event.triggerResponse();
    }

    @Override
    public void open(DaRPCServerEndpoint<Request, Response> rpcClientEndpoint) {
        System.out.print("Received new connection  ");
        // try{
        //     System.out.println(rpcClientEndpoint.getDstAddr()+" "+rpcClientEndpoint.getSrcAddr());
        //     String clientIP = ((InetSocketAddress)rpcClientEndpoint.getDstAddr()).getHostName();
        //     InetSocketAddress clientSocketAddress = new InetSocketAddress(clientIP,1921);
        //     System.out.println("Creating Endpoint");
        //     var invalidationClientEp = invalidationClientGroup.createEndpoint();
        //     invalidationClientEp.connect(clientSocketAddress, 100);
        //     System.out.println("accepted");
        //     if(invalidationClientEp.isConnected()){
        //         invalidationClientEpStreams.put(rpcClientEndpoint.getDstAddr(),invalidationClientEp.createStream());
        //         invalidationClientEps.put(rpcClientEndpoint.getDstAddr(),invalidationClientEp);
        //     }
        //     else{
        //         System.out.println("Not able to Connect to client for invalidation");
        //     }
        // }
        // catch(Exception e){
        //     e.printStackTrace();
        // }
    }

    @Override
    public void close(DaRPCServerEndpoint<Request, Response> rpcClientEndpoint) {
        System.out.println("Closing Connection");
        // TODO : Remove the connection from the static lists stores in Network Handler
        
    }
    
}
